# This is a configuration template for a single worker instance, and is
# used by Dockerfile-workers.
# Values will be change depending on whichever workers are selected when
# running that image.

worker_app: "{{ app }}"
worker_name: "{{ name }}"

worker_listeners:
  - type: http
{% if using_unix_sockets %}
    path: "/run/worker.{{ port }}"
{% else %}
    port: {{ port }}
{% endif %}
{% if listener_resources %}
    resources:
      - names:
{%- for resource in listener_resources %}
        - {{ resource }}
{%- endfor %}
{% endif %}

worker_log_config: {{ worker_log_config_filepath }}

{{ worker_extra_conf }}
